@@ -8,7 +8,7 @@
  *	console.c
  *
  * This module implements the console io functions
- *	'void con_init(void)'
+ *	'long con_init(long)'
  *	'void con_write(struct tty_queue * queue)'
  * Hopefully this will be a rather complete VT102 implementation.
  *
@@ -40,21 +40,11 @@
 #include <asm/system.h>
 #include <asm/segment.h>
 
-#include <string.h>
+#include <linux/string.h>
 #include <errno.h>
 
-#define DEF_TERMIOS \
-(struct termios) { \
-	ICRNL, \
-	OPOST | ONLCR, \
-	0, \
-	IXON | ISIG | ICANON | ECHO | ECHOCTL | ECHOKE, \
-	0, \
-	INIT_C_CC \
-}
-
-static void blank_screen(void);
-static void unblank_screen(void);
+#include <sys/kd.h>
+#include "vt_kern.h"
 
 /*
  * These are set up by the setup-routine at boot-time:
@@ -77,13 +67,17 @@ static void unblank_screen(void);
 
 #define NPAR 16
 
-int NR_CONSOLES = 0;
-
+extern void vt_init(void);
 extern void keyboard_interrupt(void);
 extern void set_leds(void);
 extern unsigned char kapplic;
+extern unsigned char ckmode;
+extern unsigned char krepeat;
 extern unsigned char kleds;
 extern unsigned char kmode;
+extern unsigned char kraw;
+extern unsigned char ke0;
+extern unsigned char lfnlmode;
 
 unsigned long	video_num_columns;		/* Number of text columns	*/
 unsigned long	video_num_lines;		/* Number of test lines		*/
@@ -95,38 +89,65 @@ static unsigned long	video_size_row;		/* Bytes per row		*/
 static unsigned char	video_page;		/* Initial video page		*/
 static unsigned short	video_port_reg;		/* Video register select port	*/
 static unsigned short	video_port_val;		/* Video register value port	*/
-static int can_do_colour = 0;
+static int can_do_color = 0;
 
 static struct {
-	unsigned short	vc_video_erase_char;	
-	unsigned char	vc_attr;
-	unsigned char	vc_def_attr;
-	int		vc_bold_attr;
-	unsigned long	vc_ques;
-	unsigned long	vc_state;
-	char *	 	vc_restate;
-	unsigned long	vc_checkin;
+	unsigned short	vc_video_erase_char;	/* Background erase character */
+	unsigned char	vc_attr;		/* Current attributes */
+	unsigned char	vc_def_color;		/* Default colors */
+	unsigned char	vc_color;		/* Foreground & background */
+	unsigned char	vc_s_color;		/* Saved foreground & background */
+	unsigned char	vc_ulcolor;		/* Colour for underline mode */
+	unsigned char	vc_halfcolor;		/* Colour for half intensity mode */
 	unsigned long	vc_origin;		/* Used for EGA/VGA fast scroll	*/
 	unsigned long	vc_scr_end;		/* Used for EGA/VGA fast scroll	*/
 	unsigned long	vc_pos;
 	unsigned long	vc_x,vc_y;
 	unsigned long	vc_top,vc_bottom;
+	unsigned long	vc_state;
 	unsigned long	vc_npar,vc_par[NPAR];
 	unsigned long	vc_video_mem_start;	/* Start of video RAM		*/
 	unsigned long	vc_video_mem_end;	/* End of video RAM (sort of)	*/
-	unsigned int	vc_saved_x;
-	unsigned int	vc_saved_y;
-	unsigned int	vc_iscolor;
-	unsigned char	vc_kbdapplic;
-	unsigned char   vc_kbdleds;
+	unsigned long	vc_saved_x;
+	unsigned long	vc_saved_y;
+	/* mode flags */
+	unsigned long	vc_kbdapplic	: 1;	/* Application keyboard */
+	unsigned long	vc_charset	: 1;	/* Character set G0 / G1 */
+	unsigned long	vc_s_charset	: 1;	/* Saved character set */
+	unsigned long	vc_decckm	: 1;	/* Cursor Keys Mode */
+	unsigned long	vc_decscnm	: 1;	/* Screen Mode */
+	unsigned long	vc_decom	: 1;	/* Origin Mode */
+	unsigned long	vc_decawm	: 1;	/* Autowrap Mode */
+	unsigned long	vc_decarm	: 1;	/* Autorepeat Mode */
+	unsigned long	vc_deccm	: 1;	/* Cursor Visible */
+	unsigned long	vc_decim	: 1;	/* Insert Mode */
+	unsigned long	vc_lnm		: 1;	/* Line feed New line Mode */
+	/* attribute flags */
+	unsigned long	vc_intensity	: 2;	/* 0=half-bright, 1=normal, 2=bold */
+	unsigned long	vc_underline	: 1;
+	unsigned long	vc_blink	: 1;
+	unsigned long	vc_reverse	: 1;
+	unsigned long	vc_s_intensity	: 2;	/* saved rendition */
+	unsigned long	vc_s_underline	: 1;
+	unsigned long	vc_s_blink	: 1;
+	unsigned long	vc_s_reverse	: 1;
+	/* misc */
+	unsigned long	vc_ques		: 1;
+	unsigned long	vc_need_wrap	: 1;
+	unsigned long	vc_tab_stop[5];		/* Tab stops. 160 columns. */
 	unsigned char	vc_kbdmode;
 	char *		vc_translate;
-} vc_cons [MAX_CONSOLES];
+	char *	 	vc_G0_charset;
+	char *	 	vc_G1_charset;
+	char *		vc_saved_G0;
+	char *		vc_saved_G1;
+	/* additional information is in vt_kern.h */
+} vc_cons [NR_CONSOLES];
 
 #define MEM_BUFFER_SIZE (2*80*50*8) 
 
-unsigned short *vc_scrbuf[MAX_CONSOLES];
-unsigned short vc_scrmembuf[MEM_BUFFER_SIZE/2];
+unsigned short *vc_scrbuf[NR_CONSOLES];
+static unsigned short * vc_scrmembuf;
 static int console_blanked = 0;
 
 #define origin		(vc_cons[currcons].vc_origin)
@@ -137,8 +158,6 @@ static int console_blanked = 0;
 #define x		(vc_cons[currcons].vc_x)
 #define y		(vc_cons[currcons].vc_y)
 #define state		(vc_cons[currcons].vc_state)
-#define restate		(vc_cons[currcons].vc_restate)
-#define checkin		(vc_cons[currcons].vc_checkin)
 #define npar		(vc_cons[currcons].vc_npar)
 #define par		(vc_cons[currcons].vc_par)
 #define ques		(vc_cons[currcons].vc_ques)
@@ -146,14 +165,50 @@ static int console_blanked = 0;
 #define saved_x		(vc_cons[currcons].vc_saved_x)
 #define saved_y		(vc_cons[currcons].vc_saved_y)
 #define translate	(vc_cons[currcons].vc_translate)
+#define G0_charset	(vc_cons[currcons].vc_G0_charset)
+#define G1_charset	(vc_cons[currcons].vc_G1_charset)
+#define saved_G0	(vc_cons[currcons].vc_saved_G0)
+#define saved_G1	(vc_cons[currcons].vc_saved_G1)
 #define video_mem_start	(vc_cons[currcons].vc_video_mem_start)
 #define video_mem_end	(vc_cons[currcons].vc_video_mem_end)
-#define def_attr	(vc_cons[currcons].vc_def_attr)
-#define video_erase_char  (vc_cons[currcons].vc_video_erase_char)	
-#define iscolor		(vc_cons[currcons].vc_iscolor)
+#define video_erase_char (vc_cons[currcons].vc_video_erase_char)	
+#define decckm		(vc_cons[currcons].vc_decckm)
+#define decscnm		(vc_cons[currcons].vc_decscnm)
+#define decom		(vc_cons[currcons].vc_decom)
+#define decawm		(vc_cons[currcons].vc_decawm)
+#define decarm		(vc_cons[currcons].vc_decarm)
+#define deccm		(vc_cons[currcons].vc_deccm)
+#define decim		(vc_cons[currcons].vc_decim)
+#define lnm		(vc_cons[currcons].vc_lnm)
 #define kbdapplic	(vc_cons[currcons].vc_kbdapplic)
+#define need_wrap	(vc_cons[currcons].vc_need_wrap)
+#define color		(vc_cons[currcons].vc_color)
+#define s_color		(vc_cons[currcons].vc_s_color)
+#define def_color	(vc_cons[currcons].vc_def_color)
+#define	foreground	(color & 0x0f)
+#define background	(color & 0xf0)
+#define charset		(vc_cons[currcons].vc_charset)
+#define s_charset	(vc_cons[currcons].vc_s_charset)
+#define	intensity	(vc_cons[currcons].vc_intensity)
+#define	underline	(vc_cons[currcons].vc_underline)
+#define	blink		(vc_cons[currcons].vc_blink)
+#define	reverse		(vc_cons[currcons].vc_reverse)
+#define	s_intensity	(vc_cons[currcons].vc_s_intensity)
+#define	s_underline	(vc_cons[currcons].vc_s_underline)
+#define	s_blink		(vc_cons[currcons].vc_s_blink)
+#define	s_reverse	(vc_cons[currcons].vc_s_reverse)
+#define	ulcolor		(vc_cons[currcons].vc_ulcolor)
+#define	halfcolor	(vc_cons[currcons].vc_halfcolor)
 #define kbdmode		(vc_cons[currcons].vc_kbdmode)
-#define kbdleds		(vc_cons[currcons].vc_kbdleds)
+#define tab_stop	(vc_cons[currcons].vc_tab_stop)
+#define kbdraw		(vt_cons[currcons].vc_kbdraw)
+#define kbdleds		(vt_cons[currcons].vc_kbdleds)
+#define vtmode		(vt_cons[currcons].vt_mode)
+
+#define SET(mode,fg,v) \
+	(mode) = (v); \
+	if (currcons == fg_console) \
+		(fg) = (v)
 
 int blankinterval = 5*60*HZ;
 static int screen_size = 0;
@@ -161,46 +216,86 @@ static int screen_size = 0;
 static void sysbeep(void);
 
 /*
- * this is what the terminal answers to a ESC-Z or csi0c
- * query (= vt100 response).
+ * this is what the terminal answers to a ESC-Z or csi0c query.
  */
-#define RESPONSE "\033[?1;2c"
+#define VT100ID "\033[?1;2c"
+#define VT102ID "\033[?6c"
 
 static char * translations[] = {
-/* normal 7-bit ascii */
+/* 8-bit Latin-1 mapped to the PC charater set: '\0' means non-printable */
+	"\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0"
+	"\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0"
 	" !\"#$%&'()*+,-./0123456789:;<=>?"
 	"@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\\]^_"
-	"`abcdefghijklmnopqrstuvwxyz{|}~ ",
+	"`abcdefghijklmnopqrstuvwxyz{|}~\0"
+	"\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0"
+	"\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0"
+	"\040\255\233\234\376\235\174\025\376\376\246\256\252\055\376\376"
+	"\370\361\375\376\376\346\024\371\376\376\247\257\254\253\376\250"
+	"\376\376\376\376\216\217\222\200\376\220\376\376\376\376\376\376"
+	"\376\245\376\376\376\376\231\376\376\376\376\376\232\376\376\341"
+	"\205\240\203\376\204\206\221\207\212\202\210\211\215\241\214\213"
+	"\376\244\225\242\223\376\224\366\376\227\243\226\201\376\376\230",
 /* vt100 graphics */
-	" !\"#$%&'()*+,-./\333123456789:;<=>?"
-	"@ABCDEFGH\017JKLMNOPQRSTUVWXYZ[\\]^ "
-	/* '   a   b   c   d   e   f   g   h   i   j   k   l   m   n   o */
-	"\004\261\007\007\007\007\370\361\040\007\331\277\332\300\305\007"
-	/* p   q   r   s   t   u   v   w   x   y   z   {   |   }   ~     */
-	"\007\304\007\007\303\264\301\302\263\007\007\007\007\007\234 "
-	
-      /*"\004\261\007\007\007\007\370\361\007\007\275\267\326\323\327\304"
-        "\304\304\304\304\307\266\320\322\272\363\362\343\\007\234\007 " */
+	"\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0"
+	"\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0"
+	" !\"#$%&'()*+,-./0123456789:;<=>?"
+	"@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\\]^ "
+	"\004\261\007\007\007\007\370\361\007\007\275\267\326\323\327\304"
+	"\304\304\304\304\307\266\320\322\272\363\362\343\007\234\007\0"
+	"\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0"
+	"\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0"
+	"\040\255\233\234\376\235\174\025\376\376\246\256\252\055\376\376"
+	"\370\361\375\376\376\346\024\371\376\376\247\257\254\253\376\250"
+	"\376\376\376\376\216\217\222\200\376\220\376\376\376\376\376\376"
+	"\376\245\376\376\376\376\231\376\376\376\376\376\232\376\376\341"
+	"\205\240\203\376\204\206\221\207\212\202\210\211\215\241\214\213"
+	"\376\244\225\242\223\376\224\366\376\227\243\226\201\376\376\230"
 };
 
 #define NORM_TRANS (translations[0])
 #define GRAF_TRANS (translations[1])
 
-/* NOTE! gotoxy thinks x==video_num_columns is ok */
-static inline void gotoxy(int currcons, unsigned int new_x,unsigned int new_y)
+static unsigned char color_table[] = { 0, 4, 2, 6, 1, 5, 3, 7,
+				       8,12,10,14, 9,13,11,15 };
+
+/*
+ * gotoxy() must verify all boundaries, because the arguments
+ * might also be negative. If the given position is out of
+ * bounds, the cursor is placed at the nearest margin.
+ */
+static void gotoxy(int currcons, int new_x, int new_y)
 {
-	if (new_x > video_num_columns || new_y >= video_num_lines)
-		return;
-	x = new_x;
-	y = new_y;
+	int max_y;
+
+	if (new_x < 0)
+		x = 0;
+	else
+		if (new_x >= video_num_columns)
+			x = video_num_columns - 1;
+		else
+			x = new_x;
+ 	if (decom) {
+		new_y += top;
+		max_y = bottom;
+	} else
+		max_y = video_num_lines;
+	if (new_y < 0)
+		y = 0;
+	else
+		if (new_y >= max_y)
+			y = max_y - 1;
+		else
+			y = new_y;
 	pos = origin + y*video_size_row + (x<<1);
+	need_wrap = 0;
 }
 
-static inline void set_origin(int currcons)
+static void set_origin(int currcons)
 {
 	if (video_type != VIDEO_TYPE_EGAC && video_type != VIDEO_TYPE_EGAM)
 		return;
-	if (currcons != fg_console)
+	if (currcons != fg_console || vtmode == KD_GRAPHICS)
 		return;
 	cli();
 	outb_p(12, video_port_reg);
@@ -291,6 +386,7 @@ static void lf(int currcons)
 		return;
 	} else 
 		scrup(currcons,top,bottom);
+	need_wrap = 0;
 }
 
 static void ri(int currcons)
@@ -301,20 +397,31 @@ static void ri(int currcons)
 		return;
 	} else
 		scrdown(currcons,top,bottom);
+	need_wrap = 0;
 }
 
-static void cr(int currcons)
+static inline void cr(int currcons)
 {
 	pos -= x<<1;
-	x=0;
+	need_wrap = x = 0;
 }
 
-static void del(int currcons)
+static inline void bs(int currcons)
+{
+	if (x) {
+		pos -= 2;
+		x--;
+		need_wrap = 0;
+	}
+}
+
+static inline void del(int currcons)
 {
 	if (x) {
 		pos -= 2;
 		x--;
 		*(unsigned short *)pos = video_erase_char;
+		need_wrap = 0;
 	}
 }
 
@@ -329,7 +436,7 @@ static void csi_J(int currcons, int vpar)
 			start = pos;
 			break;
 		case 1:	/* erase from start to cursor */
-			count = (pos-origin)>>1;
+			count = ((pos-origin)>>1)+1;
 			start = origin;
 			break;
 		case 2: /* erase whole display */
@@ -345,6 +452,7 @@ static void csi_J(int currcons, int vpar)
 		::"c" (count),
 		"D" (start),"a" (video_erase_char)
 		:"cx","di");
+	need_wrap = 0;
 }
 
 static void csi_K(int currcons, int vpar)
@@ -354,14 +462,12 @@ static void csi_K(int currcons, int vpar)
 
 	switch (vpar) {
 		case 0:	/* erase from cursor to end of line */
-			if (x>=video_num_columns)
-				return;
 			count = video_num_columns-x;
 			start = pos;
 			break;
 		case 1:	/* erase from start of line to cursor */
 			start = pos - (x<<1);
-			count = (x<video_num_columns)?x:video_num_columns;
+			count = x+1;
 			break;
 		case 2: /* erase whole line */
 			start = pos - (x<<1);
@@ -376,66 +482,101 @@ static void csi_K(int currcons, int vpar)
 		::"c" (count),
 		"D" (start),"a" (video_erase_char)
 		:"cx","di");
+	need_wrap = 0;
 }
 
-static void csi_m(int currcons )
+/*
+ *  I hope this works. The monochrome part is untested.
+ */
+static void update_attr(int currcons)
+{
+	attr = color;
+	if (can_do_color) {
+		if (underline)
+			attr = (attr & 0xf0) | ulcolor;
+		else if (intensity == 0)
+			attr = (attr & 0xf0) | halfcolor;
+	}
+	if (reverse ^ decscnm)
+		attr = (attr & 0x88) | (((attr >> 4) | (attr << 4)) & 0x77);
+	if (blink)
+		attr ^= 0x80;
+	if (intensity == 2)
+		attr ^= 0x08;
+	if (!can_do_color) {
+		if (underline)
+			attr = (attr & 0xf8) | 0x01;
+		else if (intensity == 0)
+			attr = (attr & 0xf0) | 0x08;
+	}
+	if (decscnm)
+		video_erase_char = ((color & 0x88) | (((color >> 4) |
+			(color << 4)) & 0x77) << 8) | ' ';
+	else
+		video_erase_char = (color << 8) | ' ';
+}
+
+static void default_attr(int currcons) {
+	intensity = 1;
+	underline = 0;
+	reverse = 0;
+	blink = 0;
+	color = def_color;
+}
+
+static void csi_m(int currcons)
 {
 	int i;
-	static int conv_table[8] = { 0, 4, 2, 6, 1, 5, 3, 7 };
 
 	for (i=0;i<=npar;i++)
 		switch (par[i]) {
-			case 0: attr=def_attr;break;  /* default */
-			case 1: attr=(iscolor?attr|0x08:attr|0x0f);break;  /* bold */
-			/*case 4: attr=attr|0x01;break;*/  /* underline */
-			case 4: /* bold */ 
-			  if (!iscolor)
-			    attr |= 0x01;
-			  else
-			  { /* check if forground == background */
-			    if (vc_cons[currcons].vc_bold_attr != -1)
-			      attr = (vc_cons[currcons].vc_bold_attr&0x0f)|(0xf0&(attr));
-			    else
-			    { short newattr = (attr&0xf0)|(0xf&(~attr));
-			      attr = ((newattr&0xf)==((attr>>4)&0xf)? 
-			        (attr&0xf0)|(((attr&0xf)+1)%0xf):
-			        newattr);
-			    }    
-			  }
-			  break;
-			case 5: attr=attr|0x80;break;  /* blinking */
-			case 7: attr=(attr&0x88)|((attr<<4)&0x70)|
-                          ((attr>>4)&0x07);break;  /* negative */
-			case 22: attr=attr&0xf7;break; /* not bold */ 
-			case 24: attr=attr&0xfe;break;  /* not underline */
-			case 25: attr=attr&0x7f;break;  /* not blinking */
-			case 27: attr=def_attr;break; /* positive image */
-			case 39: attr=(attr & 0xf8)|(def_attr & 0x07); break;
-			case 49: attr=(attr & 0x8f)|(def_attr & 0x70); break;
+			case 0:	/* all attributes off */
+				default_attr(currcons);
+				break;
+			case 1:
+				intensity = 2;
+				break;
+			case 2:
+				intensity = 0;
+				break;
+			case 4:
+				underline = 1;
+				break;
+			case 5:
+				blink = 1;
+				break;
+			case 7:
+				reverse = 1;
+				break;
+			case 21:
+			case 22:
+				intensity = 1;
+				break;
+			case 24:
+				underline = 0;
+				break;
+			case 25:
+				blink = 0;
+				break;
+			case 27:
+				reverse = 0;
+				break;
+			case 39:
+				color = (def_color & 0x0f) | background;
+				break;
+			case 49:
+				color = (def_color & 0xf0) | foreground;
+				break;
 			default:
-			  if (!can_do_colour)
-			    break;
-			  iscolor = 1;
-			  if ((par[i]>=30) && (par[i]<=37))
-			    attr = (attr & 0xf8) | conv_table[par[i]-30];
-			  else  /* Background color */
-			    if ((par[i]>=40) && (par[i]<=47))
-			      attr = (attr & 0x8f) | (conv_table[par[i]-40]<<4);
-			    else
+				if (par[i] >= 30 && par[i] <= 37)
+					color = color_table[par[i]-30]
+						| background; 
+				else if (par[i] >= 40 && par[i] <= 47)
+					color = (color_table[par[i]-40]<<4)
+						| foreground;
 				break;
 		}
-}
-
-static inline void set_cursor(int currcons)
-{
-	if (currcons != fg_console)
-		return;
-	cli();
-	outb_p(14, video_port_reg);
-	outb_p(0xff&((pos-video_mem_base)>>9), video_port_val);
-	outb_p(15, video_port_reg);
-	outb_p(0xff&((pos-video_mem_base)>>1), video_port_val);
-	sti();
+	update_attr(currcons);
 }
 
 static inline void hide_cursor(int currcons)
@@ -446,19 +587,149 @@ static inline void hide_cursor(int currcons)
 	outb_p(0xff&((scr_end-video_mem_base)>>1), video_port_val);
 }
 
-static void respond(int currcons, struct tty_struct * tty)
+static inline void set_cursor(int currcons)
 {
-	char * p = RESPONSE;
-
+	if (currcons != fg_console)
+		return;
 	cli();
+	if (deccm) {
+		outb_p(14, video_port_reg);
+		outb_p(0xff&((pos-video_mem_base)>>9), video_port_val);
+		outb_p(15, video_port_reg);
+		outb_p(0xff&((pos-video_mem_base)>>1), video_port_val);
+	} else
+		hide_cursor(currcons);
+	sti();
+}
+
+static void respond_string(char * p, int currcons, struct tty_struct * tty)
+{
 	while (*p) {
 		PUTCH(*p,tty->read_q);
 		p++;
 	}
-	sti();
 	TTY_READ_FLUSH(tty);
 }
 
+static void respond_num(unsigned int n, int currcons, struct tty_struct * tty)
+{
+	char buff[3];
+	int i = 0;
+
+	do {
+		buff[i++] = (n%10)+'0';
+		n /= 10;
+	} while(n && i < 3);	/* We'll take no chances */
+	while (i--) {
+		PUTCH(buff[i],tty->read_q);
+	}
+	/* caller must flush */
+}
+
+static void cursor_report(int currcons, struct tty_struct * tty)
+{
+	PUTCH('\033', tty->read_q);
+	PUTCH('[', tty->read_q);
+	respond_num(y + (decom ? top+1 : 1), currcons, tty);
+	PUTCH(';', tty->read_q);
+	respond_num(x+1, currcons, tty);
+	PUTCH('R', tty->read_q);
+	TTY_READ_FLUSH(tty);
+}
+
+static inline void status_report(int currcons, struct tty_struct * tty)
+{
+	respond_string("\033[0n", currcons, tty);	/* Terminal ok */
+}
+
+static inline void respond_ID(int currcons, struct tty_struct * tty)
+{
+	respond_string(VT102ID, currcons, tty);
+}
+
+static void invert_screen(int currcons) {
+	unsigned char *p;
+
+	if (can_do_color)
+		for (p = (unsigned char *)origin+1; p < (unsigned char *)scr_end; p+=2)
+			*p = (*p & 0x88) | (((*p >> 4) | (*p << 4)) & 0x77);
+	else
+		for (p = (unsigned char *)origin+1; p < (unsigned char *)scr_end; p+=2)
+			*p ^= *p & 0x07 == 1 ? 0x70 : 0x77;
+}
+
+static void set_mode(int currcons, int on_off)
+{
+	int i;
+
+	for (i=0; i<=npar; i++)
+		if (ques) switch(par[i]) {	/* DEC private modes set/reset */
+			case 1:			/* Cursor keys send ^[Ox/^[[x */
+				SET(decckm,ckmode,on_off);
+				break;
+			case 3:	/* 80/132 mode switch unimplemented */
+				csi_J(currcons,2);
+				gotoxy(currcons,0,0);
+				break;
+			case 5:			/* Inverted screen on/off */
+				if (decscnm != on_off) {
+					decscnm = on_off;
+					invert_screen(currcons);
+					update_attr(currcons);
+				}
+				break;
+			case 6:			/* Origin relative/absolute */
+				decom = on_off;
+				gotoxy(currcons,0,0);
+				break;
+			case 7:			/* Autowrap on/off */
+				decawm = on_off;
+				break;
+			case 8:			/* Autorepeat on/off */
+				SET(decarm,krepeat,on_off);
+				break;
+			case 25:		/* Cursor on/off */
+				deccm = on_off;
+				set_cursor(currcons);
+				break;
+		} else switch(par[i]) {		/* ANSI modes set/reset */
+			case 4:			/* Insert Mode on/off */
+				decim = on_off;
+				break;
+			case 20:		/* Lf, Enter == CrLf/Lf */
+				SET(lnm,lfnlmode,on_off);
+				break;
+		}
+}
+
+static void setterm_command(int currcons)
+{
+	switch(par[0]) {
+		case 1:	/* set color for underline mode */
+			if (can_do_color && par[1] < 16) {
+				ulcolor = color_table[par[1]];
+				if (underline)
+					update_attr(currcons);
+			}
+			break;
+		case 2:	/* set color for half intensity mode */
+			if (can_do_color && par[1] < 16) {
+				halfcolor = color_table[par[1]];
+				if (intensity == 0)
+					update_attr(currcons);
+			}
+			break;
+		case 8:	/* store colors as defaults */
+			def_color = attr;
+			default_attr(currcons);
+			update_attr(currcons);
+			break;
+		case 9:	/* set blanking interval */
+			blankinterval = ((par[1] < 60) ? par[1] : 60) * 60 * HZ;
+			break;
+	}
+}
+
 static void insert_char(int currcons)
 {
 	unsigned int i = x;
@@ -471,11 +742,13 @@ static void insert_char(int currcons)
 		old = tmp;
 		p++;
 	}
+	need_wrap = 0;
 }
 
 static void insert_line(int currcons)
 {
 	scrdown(currcons,y,bottom);
+	need_wrap = 0;
 }
 
 static void delete_char(int currcons)
@@ -483,18 +756,18 @@ static void delete_char(int currcons)
 	unsigned int i = x;
 	unsigned short * p = (unsigned short *) pos;
 
-	if (x >= video_num_columns)
-		return;
 	while (++i < video_num_columns) {
 		*p = *(p+1);
 		p++;
 	}
 	*p = video_erase_char;
+	need_wrap = 0;
 }
 
 static void delete_line(int currcons)
 {
 	scrup(currcons,y,bottom);
+	need_wrap = 0;
 }
 
 static void csi_at(int currcons, unsigned int nr)
@@ -539,81 +812,164 @@ static void csi_M(int currcons, unsigned int nr)
 
 static void save_cur(int currcons)
 {
-	saved_x = x;
-	saved_y = y;
+	saved_x		= x;
+	saved_y		= y;
+	s_intensity	= intensity;
+	s_underline	= underline;
+	s_blink		= blink;
+	s_reverse	= reverse;
+	s_charset	= charset;
+	s_color		= color;
+	saved_G0	= G0_charset;
+	saved_G1	= G1_charset;
 }
 
 static void restore_cur(int currcons)
 {
-	gotoxy(currcons, saved_x, saved_y);
+	gotoxy(currcons,saved_x,saved_y);
+	intensity	= s_intensity;
+	underline	= s_underline;
+	blink		= s_blink;
+	reverse		= s_reverse;
+	charset		= s_charset;
+	color		= s_color;
+	G0_charset	= saved_G0;
+	G1_charset	= saved_G1;
+	translate	= charset ? G1_charset : G0_charset;
+	update_attr(currcons);
+	need_wrap = 0;
 }
 
-
 enum { ESnormal, ESesc, ESsquare, ESgetpars, ESgotpars, ESfunckey, 
-	ESsetterm, ESsetgraph, ESgraph, ESgresc, ESignore };
+	EShash, ESsetG0, ESsetG1, ESignore };
+
+static void reset_terminal(int currcons, int do_clear)
+{
+	top		= 0;
+	bottom		= video_num_lines;
+	state		= ESnormal;
+	ques		= 0;
+	translate	= NORM_TRANS;
+	G0_charset	= NORM_TRANS;
+	G1_charset	= GRAF_TRANS;
+	charset		= 0;
+	need_wrap	= 0;
+
+	decscnm		= 0;
+	decom		= 0;
+	decawm		= 1;
+	deccm		= 1;
+	decim		= 0;
+
+	if (currcons == fg_console) {
+		krepeat		= 1;
+		ckmode		= 0;
+		kapplic		= 0;
+		lfnlmode	= 0;
+		kleds		= 2;
+		kmode		= 0;
+		set_leds();
+	} else {
+		decarm		= 1;
+		decckm		= 0;
+		kbdapplic	= 0;
+		lnm		= 0;
+		kbdleds		= 2;
+		kbdmode		= 0;
+	}
+
+	default_attr(currcons);
+	update_attr(currcons);
+
+	tab_stop[0]	= 0x01010100;
+	tab_stop[1]	=
+	tab_stop[2]	=
+	tab_stop[3]	=
+	tab_stop[4]	= 0x01010101;
+
+	if (do_clear) {
+		gotoxy(currcons,0,0);
+		csi_J(currcons,2);
+		save_cur(currcons);
+	}
+}
 
 void con_write(struct tty_struct * tty)
 {
-	unsigned char c;
+	int c;
 	unsigned int currcons;
 
 	wake_up(&tty->write_q->proc_list);
 	currcons = tty - tty_table;
-	if (currcons >= MAX_CONSOLES) {
+	if (currcons >= NR_CONSOLES) {
 		printk("con_write: illegal tty\n\r");
 		return;
 	}
-	while (!EMPTY(tty->write_q)) {
-		if (tty->stopped)
-			break;
-		GETCH(tty->write_q,c);
-		if (c == 24 || c == 26)
-			state = ESnormal;
-		switch(state) {
-			case ESnormal:
-				if (c > 31 && c < 127) {
-					while (x >= video_num_columns) {
-						x -= video_num_columns;
-						pos -= video_size_row;
-						lf(currcons);
-					}
-					*(char *) pos = translate[c-32];
-					*(char *) (pos+1) = attr;
-					pos += 2;
-					x++;
-				} else if (c == 27)
-					state = ESesc;
-				else if (c == 10 || c == 11 || c == 12)
-					lf(currcons);
-				else if (c == 13)
-					cr(currcons);
-				else if (c == 127)
-					del(currcons);
-				else if (c == 8) {
-					if (x) {
-						x--;
-						pos -= 2;
-					}
-				} else if (c == 9) {
-					c = 8-(x&7);
-					x += c;
-					pos += c<<1;
-					if (x > video_num_columns) {
-						x -= video_num_columns;
-						pos -= video_size_row;
-						lf(currcons);
-					}
-					c = 9;
-				} else if (c == 7)
-					sysbeep();
-			  	else if (c == 14) {
-					checkin = 1;
-				  	translate = restate;
-			  	} else if (c == 15) {
-					translate = NORM_TRANS;
-					checkin = 0;
-			  	}
+	while (!tty->stopped &&	(c = GETCH(tty->write_q)) >= 0) {
+		if (state == ESnormal && translate[c]) {
+			if (need_wrap) {
+				cr(currcons);
+				lf(currcons);
+			}
+			if (decim)
+				insert_char(currcons);
+			c = translate[c];
+			*(char *) pos = c;
+			*(char *) (pos+1) = attr;
+			if (x == video_num_columns - 1)
+				need_wrap = decawm;
+			else {
+				x++;
+				pos+=2;
+			}
+			continue;
+		}
+
+		/*
+		 *  Control characters can be used in the _middle_
+		 *  of an escape sequence.
+		 */
+		if (c < 32 || c == 127) switch(c) {
+			case 7:
+				sysbeep();
 				break;
+			case 8:
+				bs(currcons);
+				break;
+			case 9:
+				pos -= (x << 1);
+				while (x < video_num_columns - 1) {
+					x++;
+					if (tab_stop[x >> 5] & (1 << (x & 31)))
+						break;
+				}
+				pos += (x << 1);
+				break;
+			case 10: case 11: case 12:
+				lf(currcons);
+				if (!lfnlmode)
+					break;
+			case 13:
+				cr(currcons);
+				break;
+			case 14:
+				charset = 1;
+				translate = G1_charset;
+				break;
+			case 15:
+				charset = 0;
+				translate = G0_charset;
+				break;
+			case 24: case 26:
+				state = ESnormal;
+				break;
+			case 27:
+				state = ESesc;
+				break;
+			case 127:
+				del(currcons);
+				break;
+		} else switch(state) {
 			case ESesc:
 				state = ESnormal;
 				switch (c) {
@@ -621,7 +977,8 @@ void con_write(struct tty_struct * tty)
 					state = ESsquare;
 					break;
 				  case 'E':
-					gotoxy(currcons,0,y+1);
+					cr(currcons);
+					lf(currcons);
 					break;
 				  case 'M':
 					ri(currcons);
@@ -629,8 +986,11 @@ void con_write(struct tty_struct * tty)
 				  case 'D':
 					lf(currcons);
 					break;
+				  case 'H':
+					tab_stop[x >> 5] |= (1 << (x & 31));
+					break;
 				  case 'Z':
-					respond(currcons,tty);
+					respond_ID(currcons,tty);
 					break;
 				  case '7':
 					save_cur(currcons);
@@ -639,32 +999,22 @@ void con_write(struct tty_struct * tty)
 					restore_cur(currcons);
 					break;
 				  case '(':
+					state = ESsetG0;
+					break;
 				  case ')':
-				    	state = ESsetgraph;		
+					state = ESsetG1;
 					break;
-				  case 'P':
-				    	state = ESsetterm;  
-				    	break;
 				  case '#':
-				  	state = -1;
-				  	break;  	
+					state = EShash;
+					break;
 				  case 'c':
-					tty->termios = DEF_TERMIOS;
-				  	state = ESnormal;
-				  	restate = NORM_TRANS;
-					checkin = 0;
-					top = 0;
-					bottom = video_num_lines;
-					translate = NORM_TRANS;
+					reset_terminal(currcons,1);
+					break;
 				  case '>':  /* Numeric keypad */
-				 	kbdapplic = 0;
-				 	if (currcons == fg_console)
-				 		kapplic = 0;
-				 	break;
+					SET(kbdapplic,kapplic,0);
+					break;
 				  case '=':  /* Appl. keypad */
-				 	kbdapplic = 1;
-				 	if (currcons == fg_console)
-				 		kapplic = 1;
+					SET(kbdapplic,kapplic,1);
 				 	break;
 				}	
 				break;
@@ -676,7 +1026,7 @@ void con_write(struct tty_struct * tty)
 				if (c == '[') { /* Function key */
 					state=ESfunckey;
 					break;
-				}  
+				}
 				if (ques=(c=='?'))
 					break;
 			case ESgetpars:
@@ -690,10 +1040,25 @@ void con_write(struct tty_struct * tty)
 				} else state=ESgotpars;
 			case ESgotpars:
 				state = ESnormal;
+				switch(c) {
+					case 'h':
+						set_mode(currcons,1);
+						break;
+					case 'l':
+						set_mode(currcons,0);
+						break;
+					case 'n':
+						if (!ques)
+							if (par[0] == 5)
+								status_report(currcons,tty);
+							else if (par[0] == 6)
+								cursor_report(currcons,tty);
+						break;
+				}
 				if (ques) {
 					ques = 0;
 					break;
-				}  
+				}
 				switch(c) {
 					case 'G': case '`':
 						if (par[0]) par[0]--;
@@ -747,21 +1112,35 @@ void con_write(struct tty_struct * tty)
 					case 'P':
 						csi_P(currcons,par[0]);
 						break;
-					case '@':
-						csi_at(currcons,par[0]);
+					case 'c':
+						if (!par[0])
+							respond_ID(currcons,tty);
+						break;
+					case 'g':
+						if (!par[0])
+							tab_stop[x >> 5] &= ~(1 << (x & 31));
+						else if (par[0] == 3) {
+							tab_stop[0] =
+							tab_stop[1] =
+							tab_stop[2] =
+							tab_stop[3] =
+							tab_stop[4] = 0;
+						}
 						break;
 					case 'm':
 						csi_m(currcons);
 						break;
 					case 'r':
-						if (par[0])
-							par[0]--;
+						if (!par[0])
+							par[0]++;
 						if (!par[1])
 							par[1] = video_num_lines;
+						/* Minimum allowed region is 2 lines */
 						if (par[0] < par[1] &&
 						    par[1] <= video_num_lines) {
-							top=par[0];
+							top=par[0]-1;
 							bottom=par[1];
+							gotoxy(currcons,0,0);
 						}
 						break;
 					case 's':
@@ -770,61 +1149,70 @@ void con_write(struct tty_struct * tty)
 					case 'u':
 						restore_cur(currcons);
 						break;
-					case 'l': /* blank interval */
-					case 'b': /* bold attribute */
-						if (!((npar >= 2) &&
-						  ((par[1]-13) == par[0]) && 
-						  ((par[2]-17) == par[0]))) 
-						    break;
-						if ((c=='l') && (par[0]<=60)) {
-						  blankinterval = HZ*60*par[0];
-						}
-						if (c=='b')
-						  vc_cons[currcons].vc_bold_attr
-						    = par[0];
+					case '@':
+						csi_at(currcons,par[0]);
+						break;
+					case ']': /* setterm functions */
+						setterm_command(currcons);
+						break;
 				}
 				break;
 			case ESfunckey:
 				state = ESnormal;
 				break;
-			case ESsetterm:  /* Setterm functions. */
+			case EShash:
 				state = ESnormal;
-				if (c == 'S') {
-					def_attr = attr;
-					video_erase_char = (video_erase_char&0x0ff) | (def_attr<<8);
-				} else if (c == 'L')
-					/*linewrap on*/;
-				else if (c == 'l')
-					/*linewrap off*/;
+				if (c == '8') {
+					/* DEC screen alignment test. kludge :-) */
+					video_erase_char =
+						(video_erase_char & 0xff00) | 'E';
+					csi_J(currcons, 2);
+					video_erase_char =
+						(video_erase_char & 0xff00) | ' ';
+				}
 				break;
-			case ESsetgraph:
-				if (c == '0') { 
-					if (checkin)
-						translate = GRAF_TRANS;
-					restate = GRAF_TRANS;
-				} else if (c == 'B')
-					translate = restate = NORM_TRANS;
+			case ESsetG0:
+				if (c == '0')
+					G0_charset = GRAF_TRANS;
+				else if (c == 'B')
+					G0_charset = NORM_TRANS;
+				if (charset == 0)
+					translate = G0_charset;
+				state = ESnormal;
+				break;
+			case ESsetG1:
+				if (c == '0')
+					G1_charset = GRAF_TRANS;
+				else if (c == 'B')
+					G1_charset = NORM_TRANS;
+				if (charset == 1)
+					translate = G1_charset;
 				state = ESnormal;
 				break;
 			default:
 				state = ESnormal;
 		}
 	}
-	set_cursor(currcons);
 	timer_active &= ~(1<<BLANK_TIMER);
-	if (console_blanked) {
-		timer_table[BLANK_TIMER].expires = 0;
-		timer_active |= 1<<BLANK_TIMER;
-	} else if (blankinterval) {
-		timer_table[BLANK_TIMER].expires = jiffies + blankinterval;
-		timer_active |= 1<<BLANK_TIMER;
-	}
+	if (vtmode == KD_GRAPHICS)
+		return;
+	set_cursor(currcons);
+	if (currcons == fg_console)
+		if (console_blanked) {
+			timer_table[BLANK_TIMER].expires = 0;
+			timer_active |= 1<<BLANK_TIMER;
+		} else if (blankinterval) {
+			timer_table[BLANK_TIMER].expires = jiffies + blankinterval;
+			timer_active |= 1<<BLANK_TIMER;
+		}
 }
 
 void do_keyboard_interrupt(void)
 {
 	TTY_READ_FLUSH(TTY_TABLE(0));
 	timer_active &= ~(1<<BLANK_TIMER);
+	if (vt_cons[fg_console].vt_mode == KD_GRAPHICS)
+		return;
 	if (console_blanked) {
 		timer_table[BLANK_TIMER].expires = 0;
 		timer_active |= 1<<BLANK_TIMER;
@@ -845,7 +1233,7 @@ return s;
 }
 
 /*
- *  void con_init(void);
+ *  long con_init(long);
  *
  * This routine initalizes console interrupts, and does nothing
  * else. If you want the screen to clear, call tty_write with
@@ -854,21 +1242,23 @@ return s;
  * Reads the information preserved by setup.s to determine the current display
  * type and sets everything accordingly.
  */
-void con_init(void)
+long con_init(long kmem_start)
 {
 	register unsigned char a;
 	char *display_desc = "????";
 	char *display_ptr;
 	int currcons = 0;
-	long base, term;
-	long video_memory;
-	long saveterm, savebase;
+	long base;
+	int orig_x = ORIG_X;
+	int orig_y = ORIG_Y;
 
+	vc_scrmembuf = (unsigned short *) kmem_start;
 	video_num_columns = ORIG_VIDEO_COLS;
 	video_size_row = video_num_columns * 2;
 	video_num_lines = ORIG_VIDEO_LINES;
 	video_page = ORIG_VIDEO_PAGE;
-	video_erase_char = 0x0720;
+	screen_size = (video_num_lines * video_size_row);
+	kmem_start += NR_CONSOLES * screen_size;
 	timer_table[BLANK_TIMER].fn = blank_screen;
 	timer_table[BLANK_TIMER].expires = 0;
 	if (blankinterval) {
@@ -896,7 +1286,7 @@ void con_init(void)
 	}
 	else				/* If not, it is color. */
 	{
-		can_do_colour = 1;
+		can_do_color = 1;
 		video_mem_base = 0xb8000;
 		video_port_reg	= 0x3d4;
 		video_port_val	= 0x3d5;
@@ -914,7 +1304,7 @@ void con_init(void)
 		}
 	}
 	
-	/* Let the user known what kind of display driver we are using */
+	/* Let the user know what kind of display driver we are using */
 	
 	display_ptr = ((char *)video_mem_base) + video_size_row - 8;
 	while (*display_desc)
@@ -922,80 +1312,58 @@ void con_init(void)
 		*display_ptr++ = *display_desc++;
 		display_ptr++;
 	}
-	
-	savebase = video_mem_base;
-	saveterm = video_mem_term;
-	memsetw(vc_scrmembuf,video_erase_char,MEM_BUFFER_SIZE/2);
-	video_mem_base = (long)vc_scrmembuf;
-	video_mem_term = (long)&(vc_scrmembuf[MEM_BUFFER_SIZE/2]);
-	video_memory = video_mem_term - video_mem_base;
-	screen_size = (video_num_lines * video_size_row);
-	NR_CONSOLES = video_memory / screen_size;
-	if (NR_CONSOLES > MAX_CONSOLES)
-		NR_CONSOLES = MAX_CONSOLES;
-	if (!NR_CONSOLES)
-		NR_CONSOLES = 1;
-	video_memory = screen_size;
-	
-	/* Initialize the variables used for scrolling (mostly EGA/VGA)	*/
-	
-	base = origin = video_mem_start = video_mem_base;
-	term = video_mem_end = base + video_memory;
-	scr_end	= video_mem_start + screen_size;
-	top	= 0;
-	bottom	= video_num_lines;
-  	attr = 0x07;
-  	def_attr = 0x07;
-        restate = NORM_TRANS;
-        state = ESnormal;
-        checkin = 0;
-	ques = 0;
-	iscolor = 0;
-	translate = NORM_TRANS;
-	kbdleds = 2;
-	kbdmode = 0;
-	kbdapplic = 0;
-        vc_cons[0].vc_bold_attr = -1;
 
-	gotoxy(currcons,ORIG_X,ORIG_Y);
-  	for (currcons = 1; currcons<NR_CONSOLES; currcons++) {
-		vc_cons[currcons] = vc_cons[0];
-		origin = video_mem_start = (base += video_memory);
-		scr_end = origin + video_num_lines * video_size_row;
-		video_mem_end = (term += video_memory);
-		gotoxy(currcons,0,0);
+	/* Initialize the variables used for scrolling (mostly EGA/VGA)	*/
+
+	base = (long)vc_scrmembuf;
+	for (currcons = 0; currcons<NR_CONSOLES; currcons++) {
+		pos = origin = video_mem_start = base;
+		scr_end = video_mem_end = (base += screen_size);
+		vc_scrbuf[currcons] = (unsigned short *) origin;
+		vtmode		= KD_TEXT;
+		kbdraw		= 0;
+		def_color	= 0x07;   /* white */
+		ulcolor		= 0x0f;   /* bold white */
+		halfcolor	= 0x08;   /* grey */
+		reset_terminal(currcons, currcons);
 	}
-  	for (currcons = 0; currcons<NR_CONSOLES; currcons++) 
-		vc_scrbuf[currcons] = (unsigned short *)origin;
-	currcons = 0;	
-	video_mem_base = savebase;
-	video_mem_term = saveterm;
-	
+	currcons = fg_console = 0;
+
 	video_mem_start = video_mem_base;
 	video_mem_end = video_mem_term;
-	origin	= video_mem_start;
+	origin = video_mem_start;
 	scr_end	= video_mem_start + video_num_lines * video_size_row;
-	top	= 0;
-	bottom	= video_num_lines; 
-	pos=origin + y*video_size_row + (x<<1);
+	gotoxy(currcons,0,0);
+	save_cur(currcons);
+	gotoxy(currcons,orig_x,orig_y);
 	update_screen(fg_console);
+
 	set_trap_gate(0x21,&keyboard_interrupt);
 	outb_p(inb_p(0x21)&0xfd,0x21);
 	a=inb_p(0x61);
 	outb_p(a|0x80,0x61);
 	outb_p(a,0x61);
+	return kmem_start;
 }
 
 void kbdsave(int new_console)
 {
 	int currcons = fg_console;
 	kbdmode = kmode;
+	kbdraw = kraw;
 	kbdleds = kleds;
 	kbdapplic = kapplic;
+	decckm = ckmode;
+	decarm = krepeat;
+	lnm = lfnlmode;
 	currcons = new_console;
 	kmode = (kmode & 0x3F) | (kbdmode & 0xC0);
+	kraw = kbdraw;
 	kleds = kbdleds;
 	kapplic = kbdapplic;
+	ckmode = decckm;
+	krepeat = decarm;
+	lfnlmode = lnm;
 	set_leds();
 }
 
@@ -1003,9 +1371,9 @@ static void get_scrmem(int currcons)
 {
 	memcpy((void *)vc_scrbuf[fg_console],(void *)origin, screen_size);
 	video_mem_start = (unsigned long)vc_scrbuf[fg_console];
-  	origin 	= video_mem_start;
+	origin 	= video_mem_start;
 	scr_end = video_mem_end = video_mem_start+screen_size;
-  	pos = origin + y*video_size_row + (x<<1);
+	pos = origin + y*video_size_row + (x<<1);
 }
 
 static void set_scrmem(int currcons)
@@ -1086,7 +1454,7 @@ int do_screendump(int arg)
 
 	verify_area(buf,2+video_num_columns*video_num_lines);
 	currcons = get_fs_byte(buf+1);
-	if ((currcons<0) || (currcons>NR_CONSOLES))
+	if ((currcons<0) || (currcons>=NR_CONSOLES))
 		return -EIO;
 	put_fs_byte((char)(video_num_lines),buf++);	
 	put_fs_byte((char)(video_num_columns),buf++);
@@ -1105,24 +1473,21 @@ void console_print(const char * b)
 	if (currcons<0 || currcons>=NR_CONSOLES)
 		currcons = 0;
 	while (c = *(b++)) {
-		if (c == 10) {
+		if (c == 10 || c == 13 || need_wrap) {
 			cr(currcons);
-			lf(currcons);
+			if (c == 10 || need_wrap)
+				lf(currcons);
+			need_wrap = 0;
 			continue;
 		}
-		if (c == 13) {
-			cr(currcons);
-			continue;
-		}
-		while (x >= video_num_columns) {
-			x -= video_num_columns;
-			pos -= video_size_row;
-			lf(currcons);
-		}
 		*(char *) pos = c;
 		*(char *) (pos+1) = attr;
-		pos += 2;
+		if (x == video_num_columns - 1) {
+			need_wrap = 1;
+			continue;
+		}
 		x++;
+		pos+=2;
 	}
 	set_cursor(currcons);
 }
